#!/usr/bin/python3

#
# Deploy a certificate as a given kubernetes secret.
#

import argparse
import base64
import subprocess
import sys
import yaml


KUBECTL_CMD = '/snap/bin/kubectl'


def load_to_b64(path):
    with open(path, 'rb') as f:
        return base64.b64encode(f.read()).decode('utf-8').replace('\n', '')


def main(args):
    cert = load_to_b64(args.cert_path)
    key = load_to_b64(args.key_path)

    # See if the secret currently exists.
    cmd = [KUBECTL_CMD, '--namespace={}'.format(args.namespace), 'get', 'secrets/{}'.format(args.secret)]
    p = subprocess.run(cmd, stderr=subprocess.PIPE, stdout=subprocess.DEVNULL, universal_newlines=True)
    if p.returncode == 0:
        secret_exists = True
    elif p.returncode == 1 and 'NotFound' in p.stderr:
        secret_exists = False
    else:
        print('ERROR: unknown error from kubectl: {}'.format(p.stderr), file=sys.stderr)
        sys.exit(1)

    if not secret_exists:
        cmd = [
            KUBECTL_CMD, '--namespace={}'.format(args.namespace),
            'create', 'secret', 'generic', args.secret,
            ]
        subprocess.run(cmd, check=True, stdout=subprocess.DEVNULL)
        print('OK: secret created for {}/{}'.format(args.namespace, args.secret))

    cmd = [
        KUBECTL_CMD, '--namespace={}'.format(args.namespace), 'get',
        'secrets/{}'.format(args.secret), '-o', 'yaml',
        ]
    p = subprocess.run(cmd, check=True, stdout=subprocess.PIPE, universal_newlines=True)
    secret = yaml.safe_load(p.stdout)

    if 'data' not in secret:
        secret['data'] = {}

    if secret['data'].get('tls.crt', '') == cert and secret['data'].get('tls.key', '') == key:
        print('OK: no change in secret data for {}/{}'.format(args.namespace, args.secret))
        return

    secret['data']['tls.crt'] = cert
    secret['data']['tls.key'] = key

    # kubectl complains about the creationTimestamp on replace (possibly Python encoding?)
    if 'metadata' in secret:
        del secret['metadata']['creationTimestamp']

    secret_yaml = yaml.dump(secret, default_flow_style=False)

    cmd = [
        KUBECTL_CMD, '--namespace={}'.format(args.namespace), 'replace',
        'secrets/{}'.format(args.secret), '--filename', '-',
        ]
    subprocess.run(cmd, check=True, input=secret_yaml, stdout=subprocess.DEVNULL, universal_newlines=True)
    print('OK: secret data updated for {}/{}'.format(args.namespace, args.secret))


def parse_args(argv=None):
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--cert-path', required=True)
    parser.add_argument('-k', '--key-path', required=True)
    parser.add_argument('-n', '--namespace', required=True)
    parser.add_argument('-s', '--secret', required=True)
    return parser.parse_args(argv)


if __name__ == '__main__':
    main(parse_args())
