import socket
import yaml
import zlib

from os import makedirs, path, symlink, unlink, utime, walk
from re import compile, UNICODE
from shutil import rmtree
from subprocess import check_call, CalledProcessError

from charmhelpers.contrib.charmsupport import nrpe
from charmhelpers.core import hookenv
from charmhelpers.core.host import write_file
from charms import reactive
from charms.reactive import hook, when, when_not
from charmhelpers.core.templating import render


AUTOCERT_KUBERNETES_PATH = '/usr/bin/autocert-kubernetes'
AUTOCERT_STATUS_PATH = '/var/lib/nagios/autocert-status.txt'


def file_to_units(local_path, unit_path):
    # copy a file from the charm onto our unit(s)
    if local_path[-3:] == '.py':
        perms = 0o755
    else:
        perms = 0o644

    fh = open(local_path, 'r')
    write_file(
        path=unit_path,
        content=fh.read().encode(),
        owner='root',
        perms=perms,
    )


def set_workload_status(status=None, message=None):
    """ update juju status output with a specific state/message, or  """
    if status and message:
        hookenv.status_set(status, message)
    else:
        config = hookenv.config()
        num_certs = len(cert_auth_pairs())
        certs = 'certificates'
        if num_certs == 1:
            certs = 'certificate'
        suffix = ''
        service_name = sanitise(config.get('service_name'))
        if service_name:
            suffix = ' for {}'.format(service_name)
        status_message = 'Managing {} {}{}'.format(num_certs, certs, suffix)
        hookenv.log(status_message, hookenv.INFO)
        hookenv.status_set('active', status_message)


def write_cronjob():
    data = zlib.crc32(socket.gethostname().encode('utf-8'))
    random_hour = data % 24
    random_minute = data % 60

    autocert_args = []
    if reactive.is_flag_set('nrpe-external-master.available'):
        autocert_args.extend(['--status-file', AUTOCERT_STATUS_PATH])

    render('autocert-cron.j2', '/etc/cron.d/autocert',
           {'minute': random_minute, 'hour': random_hour, 'autocert_args': " ".join(autocert_args)},
           owner='root', group='root', perms=0o644)


def write_ca_cert(crt):
    ca_path = '/etc/autocert/autocert_ca_chain.crt'
    write_file(
        path=ca_path,
        content=bytes(crt, "utf-8"),
        owner='root',
        perms=0o644,
    )
    return ca_path


def pairs_from_config(label, config):
    pairs = []
    if not config:
        return pairs

    # If the config starts with a hyphen or opening square brace, it is expected
    # to be a YAML list, otherwise it is expected to be a comma separated list.
    lines = []
    if config.startswith('-') or config.startswith('['):
        for y in yaml.safe_load(config):
            lines.append(str(y))
    else:
        lines = config.split(',')

    for s in lines:
        pair = s.split('=')
        if len(pair) != 2:
            message = "Invalid {} pair '{!r}', skipping".format(label, s)
            hookenv.log(message, hookenv.WARNING)
            continue
        pairs.append([sanitise(pair[0].strip()), pair[1].strip()])

    return pairs


def cert_auth_pairs():
    pairs = []
    config = hookenv.config()
    s = config.get('cert_auth_pairs', '')
    for pair in pairs_from_config('cert auth', s):
        domain_name, auth_token = pair
        if auth_token == '':
            message = 'No auth_token found for {}, skipping'.format(domain_name)
            hookenv.log(message, hookenv.WARNING)
            continue
        pairs.append(pair)

    return pairs


def cert_additional_names():
    pairs = {}
    config = hookenv.config()
    s = config.get('cert_additional_names', '')
    for pair in pairs_from_config('cert additional name', s):
        domain_name, name_value = pair
        names = []
        for name in name_value.split(':'):
            names.append(sanitise(name))
        if len(names) == 0:
            message = 'No cert additional names found for {}, skipping'.format(domain_name)
            hookenv.log(message, hookenv.WARNING)
            continue
        pairs[domain_name] = names

    return pairs


def cert_kubernetes_names():
    pairs = {}
    config = hookenv.config()
    s = config.get('cert_kubernetes_names', '')
    for pair in pairs_from_config('cert kubernetes name', s):
        domain_name, secret_value = pair
        secret_name = secret_value.split('/')
        if len(secret_name) != 2:
            message = 'Invalid cert kubernetes value for {}, skipping'.format(domain_name)
            hookenv.log(message, hookenv.WARNING)
            continue
        pairs[domain_name] = secret_name

    return pairs


def link_additional_names(directory, prefix, domain_name, suffix, names):
    # TODO(jsing): Consider cleaning up other links that refer to this
    # same target, since they may have been removed from the config.
    for name in names:
        target = prefix + domain_name + suffix
        link_path = path.join(directory, prefix + name + suffix)
        try:
            unlink(link_path)
        except FileNotFoundError:
            pass
        symlink(target, link_path)


def write_service_config():
    config = hookenv.config()
    autocert_host = config.get('autocert_host')
    autocert_ca_cert = config.get('autocert_ca_cert')
    if autocert_ca_cert == '':
        autocert_ca_path = None
    else:
        autocert_ca_path = write_ca_cert(autocert_ca_cert)

    service_name = sanitise(config.get('service_name'))

    service_action = sanitise(config.get('service_action'))
    service_action_cmd = config.get('service_action_cmd')
    service_test_cmd = config.get('service_test_cmd')
    if service_name == '':
        msg = 'Unable to initialise unnamed service'
        hookenv.log(msg, hookenv.ERROR)
        set_workload_status('blocked', msg)
        return False

    if service_action == '':
        service_action = None
    if service_action_cmd == '':
        service_action_cmd = None
    if service_test_cmd == '':
        service_test_cmd = None

    if service_action and service_action_cmd:
        msg = 'Only one of service_action and service_action_cmd can be specified'
        hookenv.log(msg, hookenv.ERROR)
        set_workload_status('blocked', msg)
        return False

    if service_action is None and service_action_cmd is None:
        msg = 'Neither service_action or service_action_cmd have been specified'
        hookenv.log(msg, hookenv.ERROR)
        set_workload_status('blocked', msg)
        return False

    target_filename = "/etc/autocert/" + service_name + ".ini"

    render('service-ini.j2', target_filename,
           {'autocert_host': autocert_host,
            'autocert_ca_path': autocert_ca_path,
            'service_action': service_action,
            'service_action_cmd': service_action_cmd,
            'service_name': service_name,
            'service_test_cmd': service_test_cmd},
           owner='root', group='root', perms=0o640)
    return True


def kubernetes_cert_action_cmd(cert_path, key_path, namespace, secret):
    return '{} --cert-path={} --key-path={} --namespace={} --secret={}'.format(
        AUTOCERT_KUBERNETES_PATH, cert_path, key_path, namespace, secret)


def write_cert_config():
    config = hookenv.config()
    service_name = sanitise(config.get('service_name'))
    service_dir = '/etc/autocert/{}'.format(service_name)

    # clean up previous config before rewriting fresh config from valid YAML
    if path.isdir(service_dir):
        # if the service directory already exists (i.e. skip if first run)
        for (dirpath, dirnames, filenames) in walk(service_dir):
            break

        for f in filenames:
            if not f.endswith('.ini'):
                continue
            unlink("{}/{}".format(service_dir, f))

    dir_certs = config.get('dir_certs', '')
    if not path.isabs(dir_certs):
        msg = 'dir_certs is not an absolute path'
        hookenv.log(msg, hookenv.ERROR)
        set_workload_status('blocked', msg)
        return False
    dir_certs = path.abspath(dir_certs)
    makedirs(dir_certs, mode=0o700, exist_ok=True)

    dir_keys = config.get('dir_keys', '')
    if not path.isabs(dir_keys):
        msg = 'dir_keys is not an absolute path'
        hookenv.log(msg, hookenv.ERROR)
        set_workload_status('blocked', msg)
        return False
    dir_keys = path.abspath(dir_keys)
    makedirs(dir_keys, mode=0o700, exist_ok=True)

    chain_required = config.get('chain_required')
    filename_prefix = config.get('filename_prefix', '')
    suffix_cert = config.get('suffix_cert', '')
    suffix_chain = config.get('suffix_chain', '')
    suffix_key = config.get('suffix_key', '')

    additional_names = cert_additional_names()
    kubernetes_names = cert_kubernetes_names()

    # any previous config is gone, write the new version
    for cert_auth_pair in cert_auth_pairs():
        [domain_name, auth_token] = cert_auth_pair

        path_to_cert = path.join(dir_certs, filename_prefix + domain_name + suffix_cert)
        path_to_chain = path.join(dir_certs, filename_prefix + domain_name + suffix_chain)
        path_to_key = path.join(dir_keys, filename_prefix + domain_name + suffix_key)

        if not chain_required:
            path_to_chain = None

        cert_action_cmd = None
        if domain_name in kubernetes_names:
            [namespace, secret] = kubernetes_names[domain_name]
            cert_action_cmd = kubernetes_cert_action_cmd(path_to_cert, path_to_key, namespace, secret)

        target_filename = path.join("/etc/autocert/", service_name, domain_name + ".ini")

        render('cert-ini.j2', target_filename,
               {'domain_name': domain_name,
                'auth_token': auth_token,
                'cert_action_cmd': cert_action_cmd,
                'path_to_key': path_to_key,
                'path_to_cert': path_to_cert,
                'path_to_chain': path_to_chain},
               owner='root', group='root', perms=0o640)

        names = additional_names.get(domain_name, [])
        link_additional_names(dir_certs, filename_prefix, domain_name, suffix_cert, names)
        link_additional_names(dir_certs, filename_prefix, domain_name, suffix_chain, names)
        link_additional_names(dir_keys, filename_prefix, domain_name, suffix_key, names)

    return True


def sanitise(service_name):
    # try to catch malicious values, e.g. '../../home/$USER/.config'
    pattern = compile('[^a-zA-z0-9@._-]+', UNICODE)
    safe_service_name = pattern.sub('', service_name)
    return safe_service_name


def run_autocert():
    hookenv.log('Running autocert...', hookenv.INFO)
    cmd = ['/usr/bin/run-one', '/usr/sbin/autocert']
    if reactive.is_flag_set('nrpe-external-master.available'):
        cmd.extend(['--status-file', AUTOCERT_STATUS_PATH])
    try:
        check_call(cmd)
        hookenv.log('Successfully ran autocert', hookenv.INFO)
        return True
    except CalledProcessError as e:
        hookenv.log('Failed to run autocert: {}'.format(e), hookenv.ERROR)
        set_workload_status('blocked', 'failed to run autocert')
        return False


def init_service():
    if not write_service_config():
        return False
    if not write_cert_config():
        return False
    return run_autocert()


def remove_old_service(service_name):
    if service_name != '':
        dir_name = path.isdir("/etc/autocert/" + service_name)
        if path.isdir(dir_name):
            hookenv.log(
                'Removing config for previously-configured service "{}" in {}'.format(
                    service_name, dir_name),
                hookenv.INFO,
            )
            rmtree(dir_name)
        else:
            hookenv.log(
                'Previous config referenced service "{}" in {}, but this directory no longer exists'.format(
                    service_name, dir_name),
                hookenv.WARNING,
            )
    else:
        hookenv.log('Unable to remove previous service "{!r}"'.format(service_name))


@when('apt.installed.autocert')
@when_not('autocert.installed')
def install_autocert():
    file_to_units(
        'files/autocert_kubernetes.py',
        AUTOCERT_KUBERNETES_PATH,
    )
    reactive.set_flag('autocert.installed')


@when('apt.installed.autocert')
@when_not('autocert.configured')
def configure_autocert():
    if init_service():
        reactive.set_flag('autocert.configured')
        set_workload_status()
        write_cronjob()


@when('config.changed.service_name')
def maybe_remove_old_service():
    config = hookenv.config()
    old_service_name = config.previous('service_name')
    if old_service_name:
        old_service_name = sanitise(old_service_name)
        remove_old_service(old_service_name)


@when('config.changed')
def uninitialise():
    reactive.clear_flag('autocert.configured')


@when('nrpe-external-master.available')
def configure_nagios(nagios):
    if hookenv.hook_name() == 'update-status':
        return
    # Ask charmhelpers.contrib.charmsupport's nrpe to work out our hostname
    hostname = nrpe.get_nagios_hostname()
    nrpe_setup = nrpe.NRPE(hostname=hostname, primary=False)
    # install nagios support files
    file_to_units(
        'files/check_autocert_status.py',
        '/usr/local/lib/nagios/plugins/check_autocert_status.py',
    )
    if not path.exists(AUTOCERT_STATUS_PATH):
        with open(AUTOCERT_STATUS_PATH, 'a'):
            utime(AUTOCERT_STATUS_PATH, None)  # poor man's touch(1), but works with 12.04's python3
    # use charmhelpers to create the check
    nrpe_setup.add_check(
        'autocert',
        'Verify autocert is working',
        '/usr/local/lib/nagios/plugins/check_autocert_status.py {}'.format(AUTOCERT_STATUS_PATH),
    )
    nrpe_setup.write()

    write_cronjob()


@hook('upgrade-charm')
def upgrade_charm():
    reactive.clear_flag('autocert.installed')
    uninitialise()
