#!/usr/bin/env python3

import amulet
import unittest
from base64 import b64encode
from OpenSSL import crypto


def makeRSAKey(bits=2048):
    # generate and return an RSA public key
    k = crypto.PKey()
    k.generate_key(crypto.TYPE_RSA, bits)
    return k


def makeCACert(ca_key):
    # generate and return a CA certificate, given an RSA key
    ca_cert = crypto.X509()
    ca_cert.set_version(2)
    # ca_cert.set_serial_number(1)

    ca_cert.get_subject().CN = "Autocert Charm Tests CA"

    ca_cert.gmtime_adj_notBefore(0)
    ca_cert.gmtime_adj_notAfter(90 * 24 * 60 * 60)  # 90 days validity more than enough for this test

    ca_cert.set_issuer(ca_cert.get_subject())

    ca_cert.add_extensions([
        crypto.X509Extension(bytes("basicConstraints", "utf-8"), True,
                             bytes("CA:TRUE, pathlen:0", "utf-8")),
        crypto.X509Extension(bytes("keyUsage", "utf-8"), True,
                             bytes("keyCertSign, cRLSign", "utf-8")),
        crypto.X509Extension(bytes("subjectKeyIdentifier", "utf-8"), False,
                             bytes("hash", "utf-8"), subject=ca_cert),
    ])
    ca_cert.add_extensions([
        crypto.X509Extension(bytes("authorityKeyIdentifier", "utf-8"), False,
                             bytes("keyid:always", "utf-8"), issuer=ca_cert)
    ])

    ca_cert.set_pubkey(ca_key)
    ca_cert.sign(ca_key, 'sha256')
    return ca_cert


def createAndSignCert(cn, k, ca_key, ca_cert):
    # create a CSR, and from that, a certificate, given a CN, a public key, and a CA key/cert pair
    csr = crypto.X509Req()

    csr.get_subject().CN = cn

    csr.set_pubkey(k)
    csr.sign(ca_key, 'sha256')

    cert = crypto.X509()
    cert.set_version(2)

    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(30 * 24 * 60 * 60)  # 30 days validity more than enough for this test

    cert.set_subject(csr.get_subject())
    cert.set_pubkey(csr.get_pubkey())
    cert.set_issuer(ca_cert.get_subject())

    cert.add_extensions([
        crypto.X509Extension(bytes("basicConstraints", "utf-8"), True,
                             bytes("CA:FALSE, pathlen:0", "utf-8")),
        crypto.X509Extension(bytes("subjectAltName", "utf-8"), True,
                             bytes("DNS:*.localhost, DNS:localhost", "utf-8")),
    ])

    cert.sign(ca_key, 'sha256')
    return cert


class TestDeployment(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.deployment = amulet.Deployment(series='xenial')
        # deploy apache2 as our parent
        cls.deployment.add('apache2')

        # deploy our own charm
        cls.deployment.add('autocert')

        # and deploy the nrpe subordinate to test nagios checks
        cls.deployment.add('nrpe')
        cls.deployment.relate('apache2:nrpe-external-master', 'nrpe:nrpe-external-master')

        try:
            cls.deployment.setup(timeout=3600)
            cls.deployment.sentry.wait()
        except amulet.helpers.TimeoutError:
            amulet.raise_status(
                amulet.SKIP,
                msg="Environment wasn't stood up in time"
            )
        except Exception:
            raise

    def test_001_first_cert_setup(self):
        """ Set up a regular old apache2 unit to serve as a dummy autocert
        service. Once it's up, we'll generate a CA key/cert pair and use that
        to sign a service key/cert pair for "localhost" which we'll drop into
        the approprate places on the apache2 unit"""

        apache2_0 = self.deployment.sentry['apache2'][0]

        # generate an RSA public key for our CA
        ca_key = makeRSAKey(4096)
        # and a certificate
        ca_cert = makeCACert(ca_key)
        # now another RSA key for our actual site
        test_key = makeRSAKey()
        # and a certificate from that key
        test_cert = createAndSignCert("localhost", test_key, ca_key, ca_cert)

        path_to_key = '/etc/ssl/private/localhost.key'
        path_to_cert = '/etc/ssl/certs/localhost.crt'
        path_to_chain = '/etc/ssl/certs/localhost_chain.crt'

        output, exit_code = apache2_0.run('echo "{}" > {}'.format(
            crypto.dump_privatekey(crypto.FILETYPE_PEM, test_key).decode("utf-8").strip(),
            path_to_key,
        ))
        self.assertEqual(exit_code, 0)

        output, exit_code = apache2_0.run('echo "{}" > {}'.format(
            crypto.dump_certificate(crypto.FILETYPE_PEM, test_cert).decode("utf-8").strip(),
            path_to_cert,
        ))
        self.assertEqual(exit_code, 0)

        output, exit_code = apache2_0.run('echo "{}" > {}'.format(
            crypto.dump_certificate(crypto.FILETYPE_PEM, ca_cert).decode("utf-8").strip(),
            path_to_chain
        ))
        self.assertEqual(exit_code, 0)

    def test_002_apache2_setup(self):
        """ Configure apache2 to actually run using the key/cert pair from the
        last step, and then copy those files into /var/www/html so that our
        autocert client can retrieve them from the service itself """

        path_to_key = '/etc/ssl/private/localhost.key'
        path_to_cert = '/etc/ssl/certs/localhost.crt'
        path_to_chain = '/etc/ssl/certs/localhost_chain.crt'

        https_vhost_config = """<IfModule mod_ssl.c>
    <VirtualHost _default_:443>
        DocumentRoot /var/www/html
        SSLEngine on
        SSLCertificateFile {}
        SSLCertificateKeyFile {}
        SSLCertificateChainFile {}
    </VirtualHost>
</IfModule>""".format(path_to_cert, path_to_key, path_to_chain)

        https_vhost_config_b64 = b64encode(bytes(https_vhost_config, "utf-8")).decode("utf-8")

        self.deployment.configure('apache2', {
            'enable_modules': 'ssl',
            'config_change_command': 'restart'
        })
        self.deployment.configure('apache2', {'vhost_https_template': https_vhost_config_b64})
        self.deployment.sentry.wait()

        apache2_0 = self.deployment.sentry['apache2'][0]
        output, exit_code = apache2_0.run('curl -Is --cacert {} https://localhost/'.format(path_to_chain))
        self.assertEqual(exit_code, 0)

        output, exit_code = apache2_0.run('cp /etc/ssl/certs/localhost.crt /var/www/html/localhost.crt')
        self.assertEqual(exit_code, 0)
        output, exit_code = apache2_0.run('cp /etc/ssl/certs/localhost_chain.crt /var/www/html/localhost.chain')
        self.assertEqual(exit_code, 0)
        output, exit_code = apache2_0.run('cp /etc/ssl/private/localhost.key /var/www/html/localhost.key')
        self.assertEqual(exit_code, 0)

    def test_003_autocert_install(self):
        """ add the autocert subordinate to the environment at this stage now
        that our setup is ready for it """

        apache2_0 = self.deployment.sentry['apache2'][0]
        chain_cert = apache2_0.file_contents('/etc/ssl/certs/localhost_chain.crt')

        self.deployment.configure('autocert', {
            'autocert_ca_cert': chain_cert,
            'autocert_host': 'localhost',
            'cert_auth_pairs': "localhost=0xDEADBEEF",
            'chain_required': True,
            'service_action': 'reload',
            'service_name': 'apache2',
            'service_test_cmd': '/usr/sbin/apachectl configtest',
        })

        self.deployment.relate('apache2:juju-info', 'autocert:juju-info')
        self.deployment.relate('autocert:nrpe-external-master', 'nrpe:nrpe-external-master')

        self.deployment.sentry.wait()

        output, exit_code = apache2_0.run('dpkg -l run-one')
        self.assertEqual(exit_code, 0)

        output, exit_code = apache2_0.run('dpkg -l autocert')
        self.assertEqual(exit_code, 0)

        output, exit_code = apache2_0.run('grep run-one /etc/cron.d/autocert')
        self.assertEqual(exit_code, 0)

    def test_004_noop_client_run(self):
        """ verify that retrieving copies of the live key/cert pair results in noop """
        # ideally we'd run the charm action on the autocert unit, but
        # amulet doesn't support running actions on subordinates
        apache2_0 = self.deployment.sentry['apache2'][0]

        output, exit_code = apache2_0.run('/usr/bin/run-one /usr/sbin/autocert > /var/lib/nagios/autocert-status.txt')
        self.assertEqual(exit_code, 0)

        autocert_output = apache2_0.file_contents('/var/lib/nagios/autocert-status.txt')
        wanted_output = "OK: service 'apache2', cert 'localhost' is present and correct\n"
        self.assertEqual(autocert_output, wanted_output)

    def test_005_create_and_serve_new_certs(self):
        """ generate a new key/cert pair and drop them into /var/www/html """

        apache2_0 = self.deployment.sentry['apache2'][0]

        # generate an RSA public key for our CA
        ca_key = makeRSAKey(4096)
        # and a certificate
        ca_cert = makeCACert(ca_key)
        # now another RSA key for our actual site
        new_test_key = makeRSAKey()
        # and a certificate from that key
        new_test_cert = createAndSignCert("localhost", new_test_key, ca_key, ca_cert)

        path_to_key = '/var/www/html/localhost.key'
        path_to_cert = '/var/www/html/localhost.crt'
        path_to_chain = '/var/www/html/localhost.chain'

        output, exit_code = apache2_0.run('echo "{}" > {}'.format(
            crypto.dump_privatekey(crypto.FILETYPE_PEM, new_test_key).decode("utf-8").strip(),
            path_to_key,
        ))
        self.assertEqual(exit_code, 0)

        output, exit_code = apache2_0.run('echo "{}" > {}'.format(
            crypto.dump_certificate(crypto.FILETYPE_PEM, new_test_cert).decode("utf-8").strip(),
            path_to_cert,
        ))
        self.assertEqual(exit_code, 0)

        output, exit_code = apache2_0.run('echo "{}" > {}'.format(
            crypto.dump_certificate(crypto.FILETYPE_PEM, ca_cert).decode("utf-8").strip(),
            path_to_chain,
        ))
        self.assertEqual(exit_code, 0)

    def test_006_newcert_client_run(self):
        """ verify that retrieving copies of the new key/cert pair results in the service being reconfigured"""
        # ideally we'd run the charm action on the autocert unit, but
        # amulet doesn't support running actions on subordinates
        apache2_0 = self.deployment.sentry['apache2'][0]

        output, exit_code = apache2_0.run('/usr/bin/run-one /usr/sbin/autocert > /var/lib/nagios/autocert-status.txt')
        self.assertEqual(exit_code, 0)

        autocert_output = apache2_0.file_contents('/var/lib/nagios/autocert-status.txt')
        wanted_output = (
            "OK: service 'apache2', cert 'localhost' was succesfully renewed.\n"
            "OK: Service 'apache2' was restarted successfully\n"
        )
        self.assertEqual(autocert_output, wanted_output)

        # we've now restarted apache2 with a new cert, so reconfigure autocert to use its new CA cert
        chain_cert = apache2_0.file_contents('/etc/ssl/certs/localhost_chain.crt')
        self.deployment.configure('autocert', {'autocert_ca_cert': chain_cert})


if __name__ == '__main__':
    unittest.main()
